
#ifndef MODULES_DPC_UTILS_INCLUDE_PROXYWRAPPER_H_
#define MODULES_DPC_UTILS_INCLUDE_PROXYWRAPPER_H_

#include "ara/com/internal/proxy/proxy_adapter.h"
#include "ara/com/internal/proxy/event_adapter.h"
#include "ara/com/internal/proxy/field_adapter.h"
#include "ara/com/internal/proxy/method_adapter.h"
#include "ara/com/crc_verification.h"
#include "ara/log/logging.h"

#include <string>

#include "SubDataWrapper.h"


template<typename ProxyType, typename EventType, typename DataType>
class SubWrapper
{
private:
	std::string _name;
	uint32_t _maxSampleCount;
	std::string _serviceInstanceId;
	ara::com::FindServiceHandle _serviceHandle;
	std::mutex _proxyMutex;
	SubDataWrapper<DataType>& _dataWrapper;
	ara::log::Logger& _log;

	virtual EventType& GetEvent() = 0;

protected:
	std::shared_ptr<ProxyType> _proxy;

	void ServiceAvailabilityCallback(ara::com::ServiceHandleContainer<typename ProxyType::HandleType> handles,
				ara::com::FindServiceHandle handler)
	{
		  for (auto it : handles) {
		        std::lock_guard<std::mutex> lock(_proxyMutex);
		        if (it.GetInstanceId().ToString() == _serviceInstanceId.c_str() && _proxy == nullptr) {
		            _proxy = std::make_shared<ProxyType>(it);
		            _log.LogInfo() << "Created Proxy for: " << _name;
		            GetEvent().SetReceiveHandler([this]() { SubWrapper<ProxyType, EventType, DataType>::Callback(); });
		            GetEvent().Subscribe(_maxSampleCount);
		        }
		    }
	}

	void Callback()
	{
	   GetEvent().GetNewSamples([this](ara::com::SamplePtr<DataType const> ptr)
       {
			_dataWrapper.Process(&ptr);
       });
	}

public:
	SubWrapper<ProxyType, EventType, DataType>(SubDataWrapper<DataType>& dataWrapper, const std::string& name, ara::log::Logger& log, std::string serviceInstanceId  = "1", uint32_t maxSampleCount = 10) :
	_dataWrapper(dataWrapper), _name(name), _serviceInstanceId(serviceInstanceId), _maxSampleCount(maxSampleCount), _log(log)
	{
		_log.LogInfo() << "Proxy for: " << _name << " created";
	}

	virtual ~SubWrapper<ProxyType, EventType, DataType>(){}

	void Init()
	{
		_log.LogInfo() << "Initializing Proxy for: " << _name;
		_serviceHandle = ProxyType::StartFindService(
		[this](ara::com::ServiceHandleContainer<typename ProxyType::HandleType> handles, ara::com::FindServiceHandle handler) {
		SubWrapper<ProxyType, EventType, DataType>::ServiceAvailabilityCallback(std::move(handles), handler);
		}, ara::core::InstanceSpecifier(_name.c_str()));

		_log.LogInfo() << "Searching for services...";
	}

	void Destroy()
	{
			_log.LogInfo() << "Destroying Proxy for: " << _name;
			auto &event = GetEvent();
			if (_proxy != nullptr && event.IsSubscribed()) {
				event.UnsetReceiveHandler();
				event.Unsubscribe();
			    }
			ProxyType::StopFindService(_serviceHandle);
		    _log.LogInfo() << "Destroyed Proxy for: " << _name;
	}
};

#endif /* MODULES_DPC_UTILS_INCLUDE_PROXYWRAPPER_H_ */
